#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "ycbcr_to_rgb565.h"
#include "vidinput.h"
#include "png.h"

static void png_write(png_structp _png,png_bytep _data,png_size_t _sz){
  size_t ret;
  ret=fwrite(_data,_sz,1,(FILE *)png_get_io_ptr(_png));
  if(ret!=1)png_error(_png,"Write Error");
}

static void png_flush(png_structp _png){
  fflush((FILE *)png_get_io_ptr(_png));
}

int image_write_png(const unsigned char *_img,int _width,int _height,FILE *_fp){
  png_structp  png;
  png_infop    info;
  png_bytep   *rows;
  int          y;
  png=png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
  if(png==NULL)return -ENOMEM;
  info=png_create_info_struct(png);
  if(info==NULL){
    png_destroy_write_struct(&png,NULL);
    return -ENOMEM;
  }
  rows=(png_bytep *)png_malloc(png,_height*png_sizeof(png_bytep));
  if(rows==NULL){
    png_destroy_write_struct(&png,&info);
    return -ENOMEM;
  }
  for(y=0;y<_height;y++)rows[y]=(png_byte *)(_img+y*_width*4);
  if(setjmp(png_jmpbuf(png))){
    png_free(png,rows);
    png_destroy_write_struct(&png,&info);
    return -EINVAL;
  }
  png_set_write_fn(png,_fp,png_write,png_flush);
  png_set_compression_level(png,Z_NO_COMPRESSION);
  png_set_IHDR(png,info,_width,_height,8,PNG_COLOR_TYPE_RGB_ALPHA,
   PNG_INTERLACE_NONE,PNG_COMPRESSION_TYPE_DEFAULT,PNG_FILTER_TYPE_DEFAULT);
  png_set_rows(png,info,rows);
  png_write_png(png,info,PNG_TRANSFORM_IDENTITY,NULL);
  png_write_end(png,info);
  png_free(png,rows);
  png_destroy_write_struct(&png,&info);
  return 0;
}


using namespace mozilla;
using namespace mozilla::gfx;

int main(int _argc,const char **_argv){
  video_input      vid;
  th_info          ti;
  th_ycbcr_buffer  ycbcr;
  FILE            *fin;
  FILE            *fout;
  int              out_width;
  int              out_height;
  PRUint16        *rgb565;
  PRUint8         *rgb32;
  int              x;
  int              y;
  if(_argc!=5){
    fprintf(stderr,"Usage: %s <in.ogg> <out.png> <out_width> <out_height>\n",
     _argv[0]);
    return EXIT_FAILURE;
  }
  fin=fopen(_argv[1],"rb");
  if(fin==NULL){
    fprintf(stderr,"Could not open '%s' for reading.\n",_argv[1]);
    return EXIT_FAILURE;
  }
  if(video_input_open(&vid,fin)<0){
    fprintf(stderr,"Error reading video info from '%s'.\n",_argv[1]);
    return EXIT_FAILURE;
  }
  out_width=atoi(_argv[3]);
  out_height=atoi(_argv[4]);
  if(out_width<=0||out_height<=0){
    fprintf(stderr,"Invalid dimensions: %sx%s\n",_argv[3],_argv[4]);
    return EXIT_FAILURE;
  }
  video_input_get_info(&vid,&ti);
  if(video_input_fetch_frame(&vid,ycbcr,NULL)<0){
    fprintf(stderr,"Error reading first frae from '%s'.\n",_argv[1]);
    return EXIT_FAILURE;
  }
  rgb565=(PRUint16 *)malloc(out_width*out_height*sizeof(*rgb565));
  ScaleYCbCrToRGB565(ycbcr[0].data,
                     ycbcr[1].data,
                     ycbcr[2].data,
                     (PRUint8 *)rgb565,
                     ti.pic_x,
                     ti.pic_y,
                     ti.pic_width,
                     ti.pic_height,
                     out_width,
                     out_height,
                     ycbcr[0].stride,
                     ycbcr[1].stride,
                     out_width*sizeof(*rgb565),
                     ti.pixel_fmt==TH_PF_444?YV24
                      :ti.pixel_fmt==TH_PF_422?YV16
                      :YV12,
                     FILTER_BILINEAR);
  video_input_close(&vid);
  /*Convert back to 32-bit RGB for libpng.*/
  rgb32=(PRUint8 *)malloc(out_width*out_height*4*sizeof(*rgb32));
  for(y=0;y<out_height;y++)for(x=0;x<out_width;x++){
    rgb32[y*out_width+x<<2|0]=(rgb565[y*out_width+x]&0xF800)>>8;
    rgb32[y*out_width+x<<2|1]=(rgb565[y*out_width+x]&0x07E0)>>3;
    rgb32[y*out_width+x<<2|2]=(rgb565[y*out_width+x]&0x001F)<<3;
    rgb32[y*out_width+x<<2|3]=0xFF;
  }
  free(rgb565);
  fout=fopen(_argv[2],"wb");
  if(fout==NULL){
    fprintf(stderr,"Could not open '%s' for writing.\n",_argv[2]);
    return EXIT_FAILURE;
  }
  if(image_write_png(rgb32,out_width,out_height,fout)<0){
    fprintf(stderr,"Error writing image to '%s'.\n",_argv[2]);
    return EXIT_FAILURE;
  }
  free(rgb32);
  fclose(fout);
  return EXIT_SUCCESS;
}
